from datetime import datetime, timedelta, timezone
import pytz
import time
import calendar


def timer_decorator(func):
    """
    @description: 计时装饰器，测量函数执行时间
    @param: func: 被装饰的函数
    @return: wrapper: 装饰器函数
    @usage:
    @timer_decorator
    def test():
    print("Hello, world!")
    time.sleep(1)
    print("Goodbye, world!")
    test()
    # 输出：
    # 函数 test 执行耗时: 1.000000 秒
    """

    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        print(f"函数 {func.__name__} 执行耗时: {end_time - start_time:.6f} 秒")
        return result

    return wrapper


def format_date_to_year_day(date):
    """
    @description: 将日期格式化为年和年中的第几天，即YYYYJJJ格式，其中JJJ为年中的第几天，范围为001-366，
    例如2022001表示2022年的第一天，2022366表示2022年的最后一天。
    @param {str, datetime.datetime, int} date: 日期，可以是datetime.datetime对象，也可以是字符串，格式为YYYYMMDD或YYYYJJJ。
    @return {int}: 年和年中的第几天，格式为YYYYJJJ
    @example:
    >>> format_date_to_year_day("2022001")
    2022001
    >>> format_date_to_year_day(2022001)
    2022001
    >>> format_date_to_year_day(datetime(2022, 1, 1))
    2022001
    >>> format_date_to_year_day(20220101)
    2022001
    """
    format_date = date
    if isinstance(date, datetime):
        format_date = int(date.strftime("%Y%j"))  # 获取当前时间的年和年中的第几天
    elif isinstance(date, str):
        if len(date) == 7:
            format_date = int(date)
        elif len(date) == 8:
            format_date = int(datetime.strptime(date, "%Y%m%d").strftime("%Y%j"))
        else:
            print("date format error")
    elif isinstance(date, int):
        format_date = date
    else:
        print("date format error")
    return format_date


def convert_julian_regular_date(julian_date):
    # 将Julian日期转换为正常的日期
    year = int(str(julian_date)[:4])  # 提取年份
    day_of_year = int(str(julian_date)[4:])  # 提取一年中的第几天
    # 创建日期对象
    date = datetime(year=year, month=1, day=1)  # 设置为当年的第一天
    date += timedelta(days=day_of_year - 1)  # 增加相应天数
    print("转换后的日期:", date.strftime("%Y-%m-%d"))
    return date


def get_UTC_time(start_date, days=0):
    """

    Args:
        start_date: str, 如'2000-01-01 00:00:00'
        days: float,如7670.0625

    Returns:返回utc时间
    """
    # 定义起始时间
    start_time_utc = None
    if isinstance(start_date, datetime):
        start_time_utc = start_date.replace(tzinfo=timezone.utc)
    else:
        start_time_utc = datetime.strptime(
            start_date.astype(str).replace("000000000", "00"), "%Y-%m-%dT%H:%M:%S.%f"
        ).replace(tzinfo=timezone.utc)
    # 定义时间增量（7670.0625 天）
    days_increment = timedelta(days=days)
    # 计算当前时间
    current_time = start_time_utc + days_increment
    return current_time


def get_Beijing_time_from_timezone(date_with_timezone, days=0):
    """
    Args:
    @ date_with_timezone: 如'2020-12-31T01:30:00.000000000'

    Returns:返回北京时间
    """
    # 将UTC时间转换为北京时间
    utc_timezone = pytz.timezone("UTC")
    beijing_timezone = pytz.timezone("Asia/Shanghai")

    # 将带有时区信息的时间字符串转换为 datetime 对象
    datetime_obj = datetime.strptime(
        date_with_timezone.astype(str).replace("000000000", "00"),
        "%Y-%m-%dT%H:%M:%S.%f",
    )
    utc_datetime = utc_timezone.localize(datetime_obj)
    beijing_datetime = utc_datetime.astimezone(beijing_timezone)
    beijing_datetime = beijing_datetime + timedelta(days=days)
    return beijing_datetime.astimezone(tz=None).replace(tzinfo=None)
    # 将 datetime 对象转换为常规时间字符串
    # formatted_time = datetime_obj.strftime('%Y-%m-%d %H:%M:%S')


def get_Beijing_time(start_date, days):
    """
    Args:
        start_date: datetime
        days: float,如7670.0625
    Returns:返回beijing时间
    """
    days_increment = timedelta(days=days)
    # 计算当前时间
    current_time = start_date + days_increment
    current_time = current_time.astimezone(tz=None).replace(
        tzinfo=None
    )  # 将时间转换为本地时区
    # print(f"当前时间：{current_time}")
    return current_time


def get_Beijing_time_from_UTC(start_date, days):
    """
    Args:
        start_date: str, 如'2000-01-01 00:00:00'
        days: float,如7670.0625
    Returns:返回beijing时间
    """
    current_time = get_UTC_time(start_date, days)
    current_time = current_time.astimezone(tz=None).replace(
        tzinfo=None
    )  # 将时间转换为本地时区
    print(f"当前时间：{current_time}")
    return current_time


def get_Beijing_time_from_UTC(utc_time):
    """
    Args:
        start_date: datetime, 如2020-12-31 01:30:00+00:00
    Returns:返回beijing时间
    """
    current_time = utc_time.astimezone(tz=None).replace(
        tzinfo=None
    )  # 将时间转换为本地时区
    print(f"当前时间：{current_time}")
    return current_time


def get_closest_date_index(datetime_list, given_time):
    """
    Args:
        datetime_list:datetime 对象的集合
        given_time:给定的时间
    Returns:
    """

    # 查找集合中最接近给定时间的元素的索引
    closest_index = min(
        range(len(datetime_list)), key=lambda i: abs(datetime_list[i] - given_time)
    )
    return closest_index


def get_closest_hour_index(datetime_list, given_time):
    """
    Args:
        datetime_list:datetime 对象的集合
        given_time:给定的时间
    Returns:
    """
    # 查找集合中最接近给定时间的元素的索引
    closest_day_index = min(
        range(len(datetime_list)), key=lambda i: abs(datetime_list[i] - given_time)
    )
    closest_date = datetime_list[closest_day_index]
    hours = []
    for date in datetime_list:
        if (
            date.year == closest_date.year
            and date.month == closest_date.month
            and date.day == closest_date.day
        ):
            hours.append(date.hour)
    closest_hour_index = min(
        range(len(hours)), key=lambda i: abs(hours[i] - given_time.hour)
    )
    if given_time.tzinfo is not None:
        closest_hour_date = datetime(
            closest_date.year,
            closest_date.month,
            closest_date.day,
            hours[closest_hour_index],
            closest_date.minute,
            tzinfo=given_time.tzinfo,
        )
    else:
        closest_hour_date = datetime(
            closest_date.year,
            closest_date.month,
            closest_date.day,
            hours[closest_hour_index],
            closest_date.minute,
        )
    closest_index = datetime_list.index(closest_hour_date)
    return closest_index


def get_hour_of_year(date):
    """
    @description: 获取指定日期属于当年第x个小时
    @param: date: 日期
    @return: 获取指定日期属于当年第x个小时
    @usage:
    """
    # 获取当年的第一天
    start_of_year = datetime(date.year, 1, 1)
    # 计算给定日期与当年第一天之间的时间差
    time_difference = date - start_of_year
    # 将时间差转换为小时数
    hours_of_year = time_difference.total_seconds() / 3600
    # 加上当前日期的小时数
    hours_of_year += date.hour
    return int(hours_of_year)

def get_day_of_year_old(date):
    '''
    @description: 获取指定日期属于当年第x天
    @param: date: 日期
    @return: 获取指定日期属于当年第x天
    @usage:
    '''  
    start_of_year = datetime(date.year, 1, 1)    
    # 计算给定日期与当年第一天之间的时间差
    time_difference = date - start_of_year    
    # 将时间差转换为小时数
    days_of_year = int(time_difference.total_seconds() / (3600*24))+1
    # print(days_of_year) 
    return days_of_year

def get_day_of_year(date):
    '''
    @description: 获取指定日期是当年的第几天
    @param {datetime}: date: 日期
    @return: 指定日期是当年的第几天
    @usage:
    print(get_day_of_year(datetime(2021, 12, 31)))  # 365
    print(get_day_of_year(datetime(2020, 12, 31)))  # 366
    '''
    #判断是不是字符串的日期
    if isinstance(date,str):
        date=datetime.strptime(date,'%Y-%m-%d')
    return date.timetuple().tm_yday


def get_hour_of_month(date):
    '''
    @description: 获取指定日期属于当年第x个小时
    @param: date: 日期
    @return: 获取指定日期属于当年第x个小时
    @usage:
    '''  
    # 获取当年的第一天
    start_of_year = datetime(date.year, date.month, 1)    
    # 计算给定日期与当年第一天之间的时间差
    time_difference = date - start_of_year    
    # 将时间差转换为小时数
    hours_of_year = time_difference.total_seconds() / 3600    
    # 加上当前日期的小时数
    hours_of_year += date.hour    
    return int(hours_of_year)

def get_days_in_month(month, year):
    """
    @description: 获取指定月份的天数
    @param: month: 月份
    @param: year: 年份
    @return: 月份的天数
    @usage:
    print(get_days_in_month(2, 2021))  # 29
    print(get_days_in_month(4, 2021))  # 30
    """
    return calendar.monthrange(year, month)[1]



import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import BoundaryNorm

def plot_ekma_curve(x, y, values, colors_num=10, title='', x_title='VOCs', y_title='NOx', legend_title='O3', cmap='jet',
                    figsize=None, showplot=True):
    '''
    :param x: numpy 1D array
    :param y: numpy 1D array
    :param values: numpy 2D array,shape[len(x), len(y)]
    :param colors_num: 颜色个数
    :param title: 标题，default：''
    :param x_title: x轴标题,default:'VOCs'
    :param y_title: y轴标题,default:'NOx'
    :param legend_title: 图例标题,default:'O3'
    :param cmap: `matplotlib.colors.Colormap` or str or None, default: jet
        If a `.Colormap` instance, it will be returned. Otherwise, the name of
        a colormap known to Matplotlib, which will be resampled by *lut*. The
        value of None, means :rc:`image.cmap`.
    :param figsize: (float, float), default: :rc:`figure.figsize`
        Width, height in inches.
    :param showplot: 是否绘图，默认绘图；False将返回fig,用于保存
    :return: showplot=True，直接绘图，没有返回值。showplot=False,返回图表数据变量，可直接调用fig.savefig(output_file, dpi=600)进行保存图片。
    '''
    # 创建网格点
    # x_grid, NOx_grid = np.meshgrid(x, y)
    # 设定颜色分界点
    cmap_jet = plt.cm.get_cmap(cmap, colors_num)
    ticks = np.linspace(values.min(), values.max(), colors_num + 1)
    # boundaries = np.linspace(values.min(), values.max(), colors_num + 1)
    norm = BoundaryNorm(ticks, cmap_jet.N)
    # 绘制Ekman曲线
    fig = plt.figure(figsize=figsize)
    contour = plt.contourf(x, y, values, levels=ticks, cmap=cmap_jet, norm=norm)
    contour_lines = plt.contour(x, y, values, levels=contour.levels, colors='black',
                                linestyles='solid', linewidths=1)
    plt.clabel(contour_lines, inline=True, fontsize=8)
    cbar = plt.colorbar(contour, label=legend_title, ticks=ticks, boundaries=ticks)
    plt.xlabel(x_title)
    plt.ylabel(y_title)
    plt.title(title)
    # plt.grid(True)
    if showplot:
        # 显示图形
        plt.show()    
    return fig

def plot_multiple_ekma_curves(dict_data,cmap='jet',x_title='VOCs', y_title='NOx',title='',legend_title='O3',colors_num=10,panel_layout=None,
                             show_minmax=True,fig_size=None,showplot=True,value_format=None,sharex=True,sharey=True,
                             need_interpolation=False,interpolation_number=100,interpolation_method='kriging-gaussian',default_min_value=None,default_max_value=None,show_sup_title=False,font_name=None):
    '''
    :param dict_data: dict, key: title, value: [x,y,z]
    :param cmap: `matplotlib.colors.Colormap` or str or None, default: jet
        If a `.Colormap` instance, it will be returned. Otherwise, the name of
        a colormap known to Matplotlib, which will be resampled by *lut*. The
        value of None, means :rc:`image.cmap`.
    :param x_title: x轴标题,default:'VOCs'
    :param y_title: y轴标题,default:'NOx'
    :param title: 标题，default：''
    :param legend_title: 图例标题,default:'O3'
    :param panel_layout: tuple, default:None
        (rows, cols)
    :param show_minmax: 是否显示最大最小值
    :param fig_size: (float, float), default: :rc:`figure.figsize`
        Width, height in inches.
    :param showplot: 是否绘图，默认绘图；False将返回fig,用于保存
    :param value_format: 数值显示格式，默认None，显示原始数值;示例'.2f'：保留两位小数
    ：param sharex: 是否共享x轴，默认True
    ：param sharey: 是否共享y轴，默认True
    ：param need_interpolation: 是否需要插值，默认False;如果x,y为1D数组，z为2D数组，则需要插值
    ：param interpolation_number: 插值点个数，默认100
    ：param interpolation_method: 插值方法，默认'kriging-gaussian'
    ：param default_min_value: 默认最小值，默认None
    ：param default_max_value: 默认最大值，默认None
    ：param show_sup_title: 是否显示上部标题，默认False
    :return: showplot=True，直接绘图。showplot=False,返回图表数据变量，可直接调用fig.savefig(output_file, dpi=600)进行保存图片。    
    '''
    from esil.panelayout_helper import get_layout_col_row
    import string
    
    if font_name!=None:
        # 设置中文字体为系统中已安装的字体，如SimSun（宋体）、SimHei（黑体）
        plt.rcParams['font.sans-serif'] = [font_name]  # 设置中文字体为宋体 
        plt.rcParams['axes.unicode_minus'] = False  # 用于正常显示负号   
    data_types=dict_data.keys()
    plot_count=len(data_types)  
    plot_columns,plot_rows=get_layout_col_row(plot_count, panel_layout=panel_layout)   
    width,height=plot_columns*4,plot_rows*4+0.2   
    if fig_size is not None:
        width,height=fig_size
    label_list = ["("+i+")" for i in list(string.ascii_lowercase)[:plot_count]]    
    fig, axs = plt.subplots(plot_rows, plot_columns, figsize=(width,height),sharex=sharex, sharey=sharey) 
    if show_sup_title==False and plot_rows>1:
        y_titles = [y_title] *plot_rows
        for ax, row in zip(axs[:,0], y_titles):
            ax.set_ylabel(row, rotation=90, size=8)
    # 将 axs 转换为 numpy 数组（即使它是单个子图）
    axs = np.array([axs]).ravel()  
    for ax,data_type,fig_label in zip(axs.flat,data_types,label_list):
        x,y,z=dict_data[data_type]
        if x.ndim==1 and y.ndim==1 and z.ndim==2:
            need_interpolation=True       
        if need_interpolation:
            from esil.interpolation_helper import interpolate
            grid_x, grid_y, values = interpolate(x=x, y=y, data=z, interpolation_number=interpolation_number, method=interpolation_method)
        else:
            grid_x, grid_y, values=x,y,z
        ax.text(0.5, 1.07, fig_label+data_type, transform=ax.transAxes, fontsize=12, fontweight='bold',ha='center')
        # 设定颜色分界点
        cmap_jet = plt.cm.get_cmap(cmap, colors_num)
        ticks = np.linspace(values.min(), values.max(), colors_num + 1)
        norm = BoundaryNorm(ticks, cmap_jet.N)
        vmax = np.nanpercentile(values,99.5) if default_max_value == None else default_max_value #np.nanmax(grid_concentration)
        vmin =  np.nanpercentile(values,0.5)  if default_min_value == None else default_min_value#np.nanmin(grid_concentration)
        # 绘制Ekma曲线
        contour = ax.contourf(grid_x, grid_y, values, levels=ticks, cmap=cmap_jet, vmin=vmin, vmax=vmax, norm=norm)              
        contour_lines = ax.contour(grid_x, grid_y, values, levels=contour.levels, colors='black',
                                    linestyles='solid', linewidths=1)
        if value_format is not None:
            fmt=f"%{value_format}"
        else:
            fmt=None
        ax.clabel(contour_lines, inline=True, fontsize=8,fmt=fmt)#,fmt=value_format
        # ax.text(0.2, 0.15, fig_label+data_type, transform=ax.transAxes,fontsize=8, fontweight='bold', va='top')
        # 设置y轴范围
        x_min,x_max=np.min(grid_x),np.max(grid_x)
        ax.set_ylim(x_min,x_max)             
        # 设置 x 和 y 轴的 ticks 范围
        x_ticks,y_ticks = [],[]      
        interval =0.25
        x_ticks =np.arange(x_min, x_max+interval , interval)
        y_ticks=x_ticks
        ax.set_xticks(x_ticks)  # 设置 x 轴的 ticks 范围
        ax.set_yticks(y_ticks)  # 设置 y 轴的 ticks 范围   
        if show_minmax:      
            min_value, max_value,mean_value=np.min(values),np.max(values),np.mean(values)            
            if value_format is None:   
                value_format=".1e" if max_value>1e3 else ".2f" if max_value>1 else ".3f" if max_value>0.1 else ".4f" if max_value>0.01 else ".5f"              
            min_max_info='Min= {}, Max= {}, Mean={}'.format(format(min_value, value_format),format(max_value, value_format),format(mean_value, value_format))
            ax.set_xlabel(f"{'' if show_sup_title else  x_title}\n"+min_max_info)
        else:
            ax.set_xlabel(f"{'' if show_sup_title else  x_title}")
    if show_sup_title:
        fig.supxlabel(f'{x_title}',y=0.08,fontsize=15,fontweight="normal")#标签相对于图形的x位置，范围为0到1，默认为0.01，距离底部的距离为0.01，表示留出一小段空白。1表示距离顶部的距离为0
        fig.supylabel(f'{y_title}',x=0.08,fontsize=15,fontweight="normal")  
    plt.subplots_adjust(hspace=0.3)  # 调整子图之间的垂直间距
    # 添加颜色条
    cbar = plt.colorbar(contour, fraction=0.02, pad=0.04, label=f'{legend_title}',
                        ax=axs, orientation='vertical', shrink=0.7,ticks=ticks)  # 设置图例高度与图像高度相同, orientation='vertical', shrink=0.7   
    if showplot:       
        # 显示图形
        plt.show()   
    return fig   

if __name__=="__main__":
    xx = np.array([0, 0.25, 0.5, 0.75, 1, 1.25, 1.5])
    yy = np.array([0, 0.25, 0.5, 0.75, 1, 1.25, 1.5])
    zz = np.array([
        [95.1904027155239, 119.51705476234, 125.538875413249, 126.52194363351, 126.061810853617, 123.794758983251, 117.701316604246],
        [95.1844499121908, 124.143183973366, 132.550084116084, 134.747808580342, 135.048862974764, 133.446858290557, 127.92002709447],
        [94.4621439618665, 128.143106564696, 138.999476575862, 142.450491510101, 143.525611154167, 142.575775580789, 137.576921341636],
        [93.0234848645511, 131.516822536329, 144.887052792584, 149.629992422786, 151.492055391823, 151.181510853948, 146.671999345746],
        [90.8684726202446, 134.264331888267, 150.212812766248, 156.286311318398, 158.948195687734, 159.264064110034, 155.205261106798],
        [87.9971072289469, 136.385634620508, 154.976756496855, 162.419448196937, 165.8940320419, 166.823435349046, 163.176706624793],
        [84.409388690658, 137.880730733053, 159.178883984406, 168.029403058402, 172.329564454319, 173.859624570985, 170.586335899732]
    ])
    random_data = np.random.randn(7, 7)
    dict_data={}
    dict_data["臭氧EKMA曲线"]=[xx,yy,zz]
    # from esil.interpolation_helper import interpolate
    # grid_x, grid_y, values = interpolate(x=xx, y=yy, data=zz, interpolation_number=100, method='kriging-gaussian')
    # dict_data["插值后"]=[grid_x,grid_y,values]
    fig=plot_multiple_ekma_curves(dict_data,x_title="NOx",y_title="VOC",legend_title="O3",value_format=".1f",font_name='SimSun',showplot=True)
    
    # fig.savefig(savefile,dpi=300,bbox_inches='tight')
    # fig.show()
  
    # for i in range(1,6): 
    #     random_data = np.random.randn(7, 7)        
    #     dict_data[f"test{i}"]=[xx,yy,random_data+zz]  
    fig=plot_multiple_ekma_curves(dict_data,cmap='jet',x_title='VOCs', y_title='NOx',title='',legend_title='O3',colors_num=10,panel_layout=None,
                             show_minmax=True,fig_size=None,showplot=True,value_format='.1f',sharex=True,sharey=True,
                             need_interpolation=False,interpolation_number=100,interpolation_method='kriging-gaussian',default_min_value=None,default_max_value=None,show_sup_title=False)
    fig.show()
    fig.savefig("test_countour_plot.png")
    print("done")




import numpy as np
import matplotlib.pyplot as plt
import cmaps
import math
from scipy import stats
from sklearn.metrics import r2_score,mean_squared_error
from esil.panelayout_helper import get_layout_col_row
import string
from mpl_toolkits.axes_grid1 import ImageGrid
from scipy.interpolate import interp2d

def normalize_and_scale(data, scale_value):
    '''
    @description: 用于对数组data进行线性缩放的函数。它的目的是将输入数组 data 的值调整到一个新的范围 [0, scale_value]
    @param {data}: ndarray 数组，不能是数据框，数据框的话，需要调用to_numpy()转为数组
    @param {scale_value}: 缩放范围，即将数据归一化为0~scale_value之间的数
    @return {data}: 归一化后的数据
    '''
    min_value = np.min(data)
    normalized_data = data - min_value
    max_value = np.max(normalized_data)
    scaled_data = normalized_data / max_value * scale_value
    return scaled_data

def compute_density_percentages(x, y, fiSize):
    '''
    @description:将二维数据（x 和 y）进行归一化缩放，并计算每个点在一个网格中的密度。然后，它计算每个点的累积百分比
    '''
    denSta = np.zeros((fiSize, fiSize))#模拟了x*y=fiSize*fiSize的格子，待会要用来放统计每个格子里有几个点的结果
    xlength = len(x)
    xdenSta,ydenSta = normalize_and_scale(x, fiSize),normalize_and_scale(y, fiSize)  
    for count in range(xlength):
        m = int(xdenSta[count])
        n = int(ydenSta[count])
        if n >= fiSize:
            n = fiSize-1
        if m >= fiSize:
            m = fiSize-1
        denSta[m, n] = denSta[m, n] + 1
    z = np.zeros((x.shape[0], 1))
    for count in range(xlength):
        m = int(xdenSta[count])
        n = int(ydenSta[count])
        if n >= fiSize:
            n = fiSize-1
        if m >= fiSize:
            m = fiSize-1
        z[count, 0] = denSta[m, n]
    int_denSta=denSta.flatten().astype(int)
    #计算每个元素出现的频率，unique_elements为元素，tabulateDensta为元素对应下的出现频率
    unique_elements, tabulateDensta= np.unique(int_denSta, return_counts=True)
    #准备装累积百分比列表的矩阵
    percent = np.zeros(tabulateDensta.shape)
    percent[0] = 0 #把0的百分比归0
    sumPoint = x.shape[0]
    #装累积百分比查询表
    for count in range(1, tabulateDensta.shape[0]):
        percent[count] = percent[count - 1] + (unique_elements[count] * tabulateDensta[count] / sumPoint)
#傻瓜遍历找到原z值对应的百分比，更新z
    for count in range(z.shape[0]):
        for coutFindPercent in range(unique_elements.shape[0]):
            if z[count, 0] == unique_elements[coutFindPercent]:
                z[count, 0] = percent[coutFindPercent]
                break
    return z


def compute_density_percentage(x_coords, y_coords, grid_size):
    # Initialize a grid to count the number of points in each cell
    density_grid = np.zeros((grid_size, grid_size))
    num_points = len(x_coords)
    
    # Normalize and scale the coordinates to fit within the grid
    x_scaled = normalize_and_scale(x_coords, grid_size)
    y_scaled = normalize_and_scale(y_coords, grid_size)

    # Populate the density grid with point counts
    for i in range(num_points):
        x_index = int(x_scaled[i])
        y_index = int(y_scaled[i])
        x_index = min(x_index, grid_size - 1)
        y_index = min(y_index, grid_size - 1)
        density_grid[x_index, y_index] += 1

    # Initialize an array to store the density value for each point
    point_densities = np.zeros((num_points, 1))
    for i in range(num_points):
        x_index = int(x_scaled[i])
        y_index = int(y_scaled[i])
        x_index = min(x_index, grid_size - 1)
        y_index = min(y_index, grid_size - 1)
        point_densities[i, 0] = density_grid[x_index, y_index]

    # Flatten the density grid and calculate frequency of each density value
    flattened_density = density_grid.flatten().astype(int)
    unique_densities, density_counts = np.unique(flattened_density, return_counts=True)

    # Calculate cumulative percentages for each unique density value
    cumulative_percentages = np.zeros(density_counts.shape)
    cumulative_percentages[0] = 0  # Set percentage for zero density to 0
    total_points = num_points
    for i in range(1, density_counts.shape[0]):
        cumulative_percentages[i] = cumulative_percentages[i - 1] + (unique_densities[i] * density_counts[i] / total_points)

    # Update point densities with cumulative percentages
    for i in range(num_points):
        for j in range(unique_densities.shape[0]):
            if point_densities[i, 0] == unique_densities[j]:
                point_densities[i, 0] = cumulative_percentages[j]
                break

    return point_densities

def draw_density_plot(x, y, titleName, xName, yName, fiSize=100,showMNE=True):
    # x = data[:, 0]  # RSM
    # y = data[:, 1]  # CMAQ

    denSta = np.zeros((fiSize, fiSize))#这里模拟了x*y=fiSize*fiSize的格子，待会要用来放统计每个格子里有几个点的结果
    xlength = len(x)
    xdenSta = normalize_and_scale(x, fiSize)
    ydenSta = normalize_and_scale(y, fiSize)

    for count in range(xlength):
        m = int(xdenSta[count])
        n = int(ydenSta[count])
        if n >= fiSize:
            n = fiSize-1
        if m >= fiSize:
            m = fiSize-1
        denSta[m, n] = denSta[m, n] + 1

    z = np.zeros((x.shape[0], 1))
    for count in range(xlength):
        m = int(xdenSta[count])
        n = int(ydenSta[count])
        if n >= fiSize:
            n = fiSize-1
        if m >= fiSize:
            m = fiSize-1
        z[count, 0] = denSta[m, n]

    int_denSta=denSta.flatten().astype(int)
    #tabulateDensta = np.bincount(int_denSta)
    #计算每个元素出现的频率，unique_elements为元素，tabulateDensta为元素对应下的出现频率
    unique_elements, tabulateDensta= np.unique(int_denSta, return_counts=True)
    #准备装累积百分比列表的矩阵
    percent = np.zeros(tabulateDensta.shape)
    percent[0] = 0 #把0的百分比归0

    sumPoint = x.shape[0]
    #unique_tabulateDensta = np.unique(int_denSta)

    #装累积百分比查询表
    for count in range(1, tabulateDensta.shape[0]):
        percent[count] = percent[count - 1] + (unique_elements[count] * tabulateDensta[count] / sumPoint)
#傻瓜遍历找到原z值对应的百分比，更新z
    for count in range(z.shape[0]):
        for coutFindPercent in range(unique_elements.shape[0]):
            if z[count, 0] == unique_elements[coutFindPercent]:
                z[count, 0] = percent[coutFindPercent]
                break

    fig = plt.figure(figsize=(11.5, 9.8))
    plt.scatter(y, x, s=26, c=z, marker='s', cmap='jet')#这里调换了一下顺序，使得x轴表示的是CMAQ，y轴表示的是RSM
    #拟合直线 + 拟合评价
    minAll = np.floor(np.min([np.min(x), np.min(y)]) / 20) * 20
    maxAll = (np.floor(np.max([np.max(x), np.max(y)]) / 20) + 1) * 20
    #2点确定一条直线 （参考线）
    plt.plot([minAll, maxAll], [minAll, maxAll], 'b-.')
    if np.min([np.min(x), np.min(y)]) < 0:
        plt.axhline(y=0, color='k', linestyle='--')
        plt.axvline(x=0, color='k', linestyle='--')
    #2点确定一条直线 (拟合线)
    PolyfitResult = np.polyfit(x, y, 1)
    plt.plot([minAll, maxAll], [PolyfitResult[0] * minAll + PolyfitResult[1], PolyfitResult[0] * maxAll + PolyfitResult[1]], 'r-')

    #评价拟合质量
    N = x.shape[0]
    R2 = 1 - np.sum((PolyfitResult[0] * x + PolyfitResult[1] - y) ** 2) / np.sum((y - np.mean(y)) ** 2)

    Ec = np.abs(x - y) / np.abs(y)
    ETc = Ec[Ec <= 1000]
    Fc = np.abs(x - y) / np.abs(x + y)
    FTc = Fc[Fc <= 1000]

    MNE = np.sum(ETc) / (N) * 100
    MFE = np.sum(FTc) / (N) * 100
    RMSE = np.sqrt(np.mean((y - x) ** 2))#计算均方根误差
    MAE = np.mean(np.abs(y - x))

    MaxNE = np.sort(ETc)* 100
    MaxNE_95 =np.percentile(MaxNE, 95) #percentileofscore(MaxNE, 95)
    MaxFE = np.sort(FTc)* 100
    MaxFE_95 =np.percentile(MaxFE, 95)# percentileofscore(MaxFE, 95)

    plt.xlabel(xName, fontname='Times New Roman', fontweight='bold', fontsize=10)
    plt.ylabel(yName, fontname='Times New Roman', fontweight='bold', fontsize=10)
    plt.title(titleName, fontname='Times New Roman', fontweight='bold', fontsize=10)
    str = ['Y = {:0.4f} + {:0.4f}X'.format(PolyfitResult[1], PolyfitResult[0]),
           'R^2 = {:0.4f}'.format(R2),
           'NME = {:0.2f}%'.format(MNE),
           '95th MaxNE = {:0.2f}%'.format(MaxNE_95),
           'RMSE = {:0.4f}'.format(RMSE),
           'MAE = {:0.4f}'.format(MAE)]
    if not showMNE:
        str = ['Y = {:0.4f} + {:0.4f}X'.format(PolyfitResult[1], PolyfitResult[0]),
               'R^2 = {:0.4f}'.format(R2),
               'RMSE = {:0.4f}'.format(RMSE),
               'MAE = {:0.4f}'.format(MAE)]
    plt.text(0.04, 0.95, '\n'.join(str), fontname='Times New Roman', verticalalignment='top', horizontalalignment='left', fontweight='bold',
             fontsize=10,transform=plt.gca().transAxes)

    plt.xlim(minAll, maxAll)#设置x轴的数据显示范围
    plt.ylim(minAll, maxAll)#设置y轴的数据显示范围
    plt.xticks(np.arange(minAll, maxAll, 10))
    plt.yticks(np.arange(minAll, maxAll, 10))
    plt.tight_layout()

    #plt.colorbar(label='Percent of data point', fontweight='bold', fontsize=35)
    cbar = plt.colorbar(label='Percent of data point', orientation='vertical')
    # Set font properties for the colorbar label
    #cbar.ax.yaxis.label.set_font_properties({'weight': 'bold', 'size': 35})
    cbar.ax.yaxis.label.set_family('Times New Roman')
    cbar.ax.yaxis.label.set_fontweight('bold')
    cbar.ax.yaxis.label.set_fontsize(14)
    # Set the locator and formatter for the colorbar
    from matplotlib.ticker import MultipleLocator, FormatStrFormatter
    # 方法1：Set the ticks for the colorbar
    #ticks=np.arange(0, 1.05, 0.1)# Set the interval between ticks
    #cbar.set_ticks(ticks)  # Set the desired ticks
    cbar.locator = MultipleLocator(0.1)  # Set the interval between ticks
    cbar.update_ticks()
    #方法2
    '''
    cbar.locator = MultipleLocator(0.1)  # Set the interval between ticks
    cbar.update_ticks()
    # Add 0 as a tick
    ticks = cbar.get_ticks().tolist()
    if 0 not in ticks:
        ticks.append(0)
        cbar.set_ticks(sorted(ticks))
    '''
    plt.box(True)
    #plt.axis('normal')
    plt.axis('auto')
    #plt.show()
    #plt.pause(2)#加上plt.pause(0.001)，这样程序就会暂停执行一段极短的时间，然后继续执行后面的代码，而绘图界面仍然保持打开状态。
    # 关闭窗口
    #plt.gcf().canvas.mpl_disconnect(plt.gcf().canvas.manager.key_press_handler_id)
    return (RMSE, MAE, fig, MNE, MaxNE_95, R2)

def draw_density_plot_via_hist2d(x, y, titleName, xName="Values", yName="Estimated Values", nbins=150,cmap='jet'):
    
    # 计算统计变量
    slope, intercept, r_value, p_value, std_err = stats.linregress(x,y)
    rmse = np.sqrt(mean_squared_error(x,y))
    r2=r2_score(x,y)    
    normalized_bias = (x - y) / y   * 100
    normalized_bias_ordered = np.sort(normalized_bias)
    mean_normalized_bias =np.mean(normalized_bias)
    max_normalized_bias_95 =np.percentile(normalized_bias_ordered, 95) 

    
    H, xedges, yedges = np.histogram2d(x, y, bins=nbins,density=False)
    # H needs to be rotated and flipped
    H = np.rot90(H)
    H = np.flipud(H)
    # Mask zeros
    Hmasked = np.ma.masked_where(H==0,H) # Mask pixels with a value of zero
    #开始绘图
    fig,ax = plt.subplots(figsize=(4,3.5),dpi=100,facecolor="w")
    density_scatter = ax.pcolormesh(xedges, yedges, Hmasked, cmap=cmap,label="Data")
    linreg = ax.plot(x, intercept + slope*x, 'r', label='Fitted Line')

    colorbar = fig.colorbar(density_scatter,aspect=17,label="Density of Points")
    colorbar.ax.tick_params(left=True,direction="in",width=.4,labelsize=10)
    colorbar.ax.tick_params(which="minor",right=False)
    colorbar.outline.set_linewidth(.4)

    # 添加文本信息
    fontdict = {"size":14,"fontstyle":"italic","weight":"bold"}
    ax.text(-4,7.5,r'$R=$'+str(round(r_value,2)),fontdict=fontdict)
    ax.text(-4,5.8,r'$y=$'+str(round(slope,2))+'$x$'+ str(round(intercept,3)),
            fontdict=fontdict)
    ax.set_xlabel(xName)
    ax.set_ylabel(yName)
    ax.legend(loc="lower right",frameon=False,labelspacing=.4,handletextpad=.5,fontsize=10)

    # fig.savefig('plot_test\\data\\图4-2-13 散点密度图绘制示例_a.png', 
    #             bbox_inches='tight',dpi=300)
    plt.show()
    return fig,rmse,r_value,mean_normalized_bias,max_normalized_bias_95

def draw_density_plot_via_gaussian_kde(x, y, titleName, xName="Values", yName="Estimated Values", cmap='jet'):
    #TODO：未经测试
    from scipy.stats import gaussian_kde
    # 计算统计变量
    slope, intercept, r_value, p_value, std_err = stats.linregress(x,y)
    rmse = np.sqrt(mean_squared_error(x,y))
    r2=r2_score(x,y)    
    normalized_bias = (x - y) / y   * 100
    normalized_bias_ordered = np.sort(normalized_bias)
    mean_normalized_bias =np.mean(normalized_bias)
    max_normalized_bias_95 =np.percentile(normalized_bias_ordered, 95) 
    
    # Calculate point density
    xy = np.vstack([x, y])
    z = gaussian_kde(xy)(xy)
    # Sort points by density
    idx = z.argsort()
    x1, y1, z1 = x[idx], y[idx], z[idx]
    max_x_value=np.max(np.abs(np.min(x)),np.abs(np.max(x)))   
    #绘制最佳拟合线
    best_line_x = np.linspace(-max_x_value,max_x_value)
    best_line_y=best_line_x
    fig,ax = plt.subplots(figsize=(4,3.5),dpi=100,facecolor="w")
    gass_desity = ax.scatter(x, y,edgecolors = 'none',c = z,s =6,marker="s",cmap=cmaps.jet)
    bestline = ax.plot(best_line_x,best_line_y,color='k',linewidth=1,linestyle='-',label="1:1 Line")
    linreg = ax.plot(x, intercept + slope*x, 'r', label='Fitted Line')

    colorbar = fig.colorbar(gass_desity,aspect=17,label="Density of Points")
    colorbar.ax.tick_params(left=True,direction="in",width=.4,labelsize=10)
    colorbar.ax.tick_params(which="minor",right=True)
    colorbar.outline.set_linewidth(.4)
    # 添加文本信息
    fontdict = {"size":14,"fontstyle":"italic","weight":"bold"}
    ax.text(-5,7.5,r'$R=$'+str(round(r_value,2)),fontdict=fontdict)
    ax.text(-5,5.8,r'$y=$'+str(round(slope,2))+'$x$'+ str(round(intercept,3)),
            fontdict=fontdict)
    ax.set_xlabel(xName)
    ax.set_ylabel(yName)
    ax.legend(loc="lower right",frameon=False,labelspacing=.4,handletextpad=.5,fontsize=10)

    # fig.savefig('plot_test\\data\\图4-2-13 散点密度图绘制示例_b.png', 
    #             bbox_inches='tight',dpi=300)
    plt.show()
    return fig,rmse,r_value,mean_normalized_bias,max_normalized_bias_95

def draw_multiple_density_plot_with_hist2d(dict_data, xName="True Values", yName="Estimated Values", cmap='jet',panel_layout=None,density=True,nbins=150,show_plot=False):
    '''
    @description: 绘制多子图相关性散点图
    @param {dict_data}: 数据字典dict,包含了x和y的数据,数据格式为{data_type:{"x":x,"y":y}}
    '''
    plot_count=len(dict_data)
    plot_columns,plot_rows=get_layout_col_row(plot_count, panel_layout=panel_layout)   
    width,height=plot_columns*3,plot_rows*3+0.2
    fig = plt.figure(figsize=(width,height))  
    label_list = ["("+i+")" for i in list(string.ascii_lowercase)[:plot_count]]

    grid = ImageGrid(fig, 111,  # similar to subplot(111)
                    nrows_ncols=(plot_rows, plot_columns),  # creates 2x2 grid of axes
                    axes_pad=0.15,  # pad between axes in inch.
                    cbar_mode="single",
                    cbar_location="right")  
    data_types=dict_data.keys() 
    temp=np.array([[np.min(dict_data[data_type]["x"]),np.max(dict_data[data_type]["x"]),np.min(dict_data[data_type]["y"]),np.max(dict_data[data_type]["y"])]for data_type in data_types])
    min_x_value,max_x_value=np.min(temp),np.max(temp)
    for data_type,ax,label in zip(data_types,grid,label_list):        
        x = dict_data[data_type]["x"]
        y = dict_data[data_type]["y"]  
        # x = changex(x_origin, 100)
        # y = changex(y_origin, 100)    
        # 计算所需指标
        slope, intercept, r_value, p_value, std_err = stats.linregress(x,y)
        rmse = np.sqrt(mean_squared_error(x,y))
        # min_x_value,max_x_value=np.min(x),np.max(x)   
        min_y_value,max_y_value=np.min(y),np.max(y)   
        #绘制最佳拟合线
        best_line_x = np.linspace(min_x_value,max_x_value)
        best_line_y=best_line_x     
        # import matplotlib.colors as mcolors
        # # 设置归一化
        # norm = mcolors.Normalize(vmin=0, vmax=1)
        hist2d = ax.hist2d(x=x,y=y,bins=nbins,cmap=cmap,cmin=.0001,density=density)#cmin指定了颜色映射的最小值。小于这个值的计数或密度将被忽略，不会被绘制。图对象中其他部分被纯色填充. norm=norm
              
        ax.plot(best_line_x,best_line_y,color='k',linewidth=.8,alpha=.8,
                        linestyle='-',zorder=-1)
        ax.plot(x, intercept + slope*x, 'r',linewidth=.8)
        ax.grid(False)
        
        # 添加文本信息
        interval_x =round(float((max_x_value - min_x_value) / 5),2)
        interval_y =round(float((max_y_value - min_y_value) / 5),2)
        start_x=min_x_value+interval_x*0.5
        start_y=max_x_value-interval_x*0.5
        intercept_symbol="+" if intercept>0 else "-"
        ax.text(start_x,start_y,r'$y=$'+str(round(slope,3))+'$x$'+f" {intercept_symbol} "+str(abs(round(intercept,3))),fontsize=8)
        ax.text(start_x,start_y-0.5*interval_x,r'$R^2=$'+str(round(r_value**2,2)),fontsize=8)
        ax.text(start_x,start_y-interval_x,"RMSE="+str(round(rmse,2)),fontsize=8)
        ax.text(0.2, 0.15, label+data_type, transform=ax.transAxes,fontsize=8, fontweight='bold', va='top')
        
        x_ticks =np.round(np.arange(min_x_value, max_x_value, interval_x),0)
        # interval_y =round(float((max_latitude - min_latitude) / 7),2)
        # y_ticks =np.round(np.arange(min_latitude, max_latitude , interval_x),1)
        
        ax.set(xlim=(min_x_value,max_x_value),ylim=(min_x_value,max_x_value),xticks=x_ticks,
            yticks=x_ticks)
        
    fig.supxlabel(xName,y=0.05,fontsize=15,fontweight="normal")
    fig.supylabel(yName,x=.01,fontsize=15,fontweight="normal")
    # hist2d[3]： hist2d 是 ax.hist2d 函数的返回值，hist2d 是一个包含了直方图统计信息的对象。hist2d[3] 表示这个对象中的第四个元素，通常是用来表示颜色条的对象。
    # cax=grid.cbar_axes[0]：这里指定了颜色条的轴，即将颜色条放置在 grid.cbar_axes[0] 上。
    # orientation='vertical'：这个参数指定了颜色条的方向为垂直方向。
    # cbar = plt.colorbar(im, ax=ax)
    cbar = fig.colorbar(hist2d[3], cax=grid.cbar_axes[0], orientation='vertical')
    cbar.set_label(label="Number Of Point" if density==False else "Density Of Point",fontsize=11)
    cbar.ax.tick_params(left=True,labelright=True,direction="in",width=.4,labelsize=10)
    cbar.ax.tick_params(which="minor",right=False)
    cbar.ax.set_title("Counts" if density==False else "Density",fontsize=11)
    cbar.outline.set_linewidth(.4)
    if show_plot:
        plt.show()
    return fig

def draw_multiple_density_plots(dict_data, xName="True Values", yName="Estimated Values", cmap='jet',panel_layout=None,density=True,nbins=150,show_plot=False,show_subplot_label=True,show_best_line=True,show_r2=True):
    '''
    @description: 绘制多子图相关性散点图
    @param {dict_data}: 数据字典dict,包含了x和y的数据,数据格式为{data_type:{"x":x,"y":y}}
    @param {xName}: x轴名称
    @param {yName}: y轴名称
    @param {cmap}: 颜色映射
    @param {panel_layout}: 子图布局,默认为None,即自动计算布局。如果指定了panel_layout,则按照panel_layout的布局进行绘图。示例：panel_layout=(3,2)表示3列2行的子图布局。
    @param {density}: 是否显示密度图还是密度数量图
    @param {nbins}: 直方图的bin数量(即分组数量)
    @param {show_plot}: 是否显示绘图结果
    @param {show_subplot_label}: 是否显示子图标签
    @param {show_best_line}: 是否显示对角线
    @return {fig}: 绘图对象
    '''   
    plot_count=len(dict_data)
    plot_columns,plot_rows=get_layout_col_row(plot_count, panel_layout=panel_layout)   
    width,height=plot_columns*3,plot_rows*3+0.2
    fig = plt.figure(figsize=(width,height))  
    label_list = ["("+i+")" for i in list(string.ascii_lowercase)[:plot_count]]
    grid = ImageGrid(fig, 111,  # similar to subplot(111)
                    nrows_ncols=(plot_rows, plot_columns),  # creates 2x2 grid of axes
                    axes_pad=0.15,  # pad between axes in inch.
                    cbar_mode="single",
                    cbar_location="right")  
    data_types=dict_data.keys() 
    temp=np.array([[np.min(dict_data[data_type]["x"]),np.max(dict_data[data_type]["x"]),np.min(dict_data[data_type]["y"]),np.max(dict_data[data_type]["y"])]for data_type in data_types])
    min_x_value,max_x_value=np.min(temp),np.max(temp)
    for data_type,ax,label in zip(data_types,grid,label_list):        
        x = dict_data[data_type]["x"]
        y = dict_data[data_type]["y"]  
        z = compute_density_percentages(x, y, nbins)        
        # 计算所需指标
        slope, intercept, r_value, p_value, std_err = stats.linregress(x,y)
        rmse = np.sqrt(mean_squared_error(x,y))
        # rmse= np.sqrt(np.mean((y - x) ** 2))
        # min_x_value,max_x_value=np.min(x),np.max(x)   
        min_y_value,max_y_value=np.min(y),np.max(y)   
        #绘制最佳拟合线
        best_line_x = np.linspace(min_x_value,max_x_value)
        best_line_y=best_line_x             
        scatter_plots = ax.scatter(x, y, s=5, c=z, marker='o', cmap=cmap)
        # hist2d = ax.hist2d(x=x,y=y,bins=nbins,cmap=cmap,cmin=.0001,density=density)#cmin指定了颜色映射的最小值。小于这个值的计数或密度将被忽略，不会被绘制。图对象中其他部分被纯色填充. norm=norm
        if show_best_line:
            ax.plot(best_line_x,best_line_y,color='k',linewidth=.8,alpha=.8,
                            linestyle='-',zorder=-1)
        ax.plot(x, intercept + slope*x, 'r',linewidth=.8)
        ax.grid(False)        
        # 添加文本信息
        interval_x =round(float((max_x_value - min_x_value) / 5),2)
        interval_y =round(float((max_y_value - min_y_value) / 5),2)
        start_x=min_x_value+interval_x*0.5
        start_y=max_x_value-interval_x*0.5
        intercept_symbol="+" if intercept>0 else "-"
        ax.text(start_x,start_y,r'$y=$'+str(round(slope,3))+'$x$'+f" {intercept_symbol} "+str(abs(round(intercept,3))),fontsize=8)
        if show_r2:
            ax.text(start_x,start_y-0.5*interval_x,r'$R^2=$'+str(round(r_value**2,2)),fontsize=8)
        else:
            ax.text(start_x,start_y-0.5*interval_x,r'$R=$'+str(round(r_value,2)),fontsize=8)
        ax.text(start_x,start_y-interval_x,"RMSE="+str(round(rmse,2)),fontsize=8)
        if show_subplot_label:
            ax.text(0.2, 0.15, label+data_type, transform=ax.transAxes,fontsize=8, fontweight='bold', va='top')
        
        x_ticks =np.round(np.arange(min_x_value, max_x_value, interval_x),0)
        # interval_y =round(float((max_latitude - min_latitude) / 7),2)
        # y_ticks =np.round(np.arange(min_latitude, max_latitude , interval_x),1)        
        ax.set(xlim=(min_x_value,max_x_value),ylim=(min_x_value,max_x_value),xticks=x_ticks,
            yticks=x_ticks)        
    fig.supxlabel(xName,y=0.05,fontsize=15,fontweight="normal")
    fig.supylabel(yName,x=.01,fontsize=15,fontweight="normal")  
    cbar = fig.colorbar(scatter_plots, cax=grid.cbar_axes[0], orientation='vertical')
    cbar.set_label(label="Number Of Point" if density==False else "Density Of Point",fontsize=11)
    cbar.ax.tick_params(left=True,labelright=True,direction="in",width=.4,labelsize=10)
    cbar.ax.tick_params(which="minor",right=False)
    cbar.ax.set_title("Counts" if density==False else "Density",fontsize=11)
    # Set the locator and formatter for the colorbar
    from matplotlib.ticker import MultipleLocator, FormatStrFormatter
    cbar.locator = MultipleLocator(0.1)  # Set the interval between ticks
    cbar.update_ticks()
    cbar.outline.set_linewidth(.4)
    plt.tight_layout()
    if show_plot:
        plt.show()        
    return fig


def plot_line_with_error(save_file,df,x_column,y_columns,data_type_column="Metric",mean_column="Mean",min_column="Min",max_column="Max",**kwargs):       
    colors = ["#2FBE8F","#459DFF","#FF5B9B","#FFCC37"]   
    fig,ax = plt.subplots(figsize=(4,3.5),dpi=100,facecolor="w")
    for index,color in zip(y_columns,colors):
        data_selcet = df.loc[df[data_type_column]==index,:]
        # 计算误差
        y_err = [data_selcet[mean_column] - data_selcet[min_column], data_selcet[max_column] - data_selcet[mean_column]]
        ax.errorbar(x=data_selcet[x_column],y=data_selcet[mean_column],yerr=y_err,color="k",
                    linewidth=1.5,marker='o',ms=10,mew=2,mfc=color,mec='k',capsize=5,label=index)
    # ax.set_ylim(-8,30)
    # ax.set_xlim(-2,40)
    ax.set_xlabel("Day")
    ax.set_ylabel("Values Change")
    ax.legend(ncol=2,frameon=False)
    ax.set_axisbelow(True)

    fig.savefig(save_file,bbox_inches='tight',dpi=300)
    plt.show()

def plot_bar(save_file,df,data_columns,x_column,**kwargs):
    '''
    @description: 绘制柱状图
    @param {
        save_file: str, 保存图片的文件名，如'bar.png'
        df: pandas.DataFrame, 数据集
        data_columns: list, 包含需要绘制的数据列名称的列表，如['CO2','CH4','N2O']
        x_column: str, x轴数据列名，如'City'        
        x_title: str, x轴标题，默认为x_column
        y_title: str, y轴标题，默认为None
        title: str, 图形标题，默认为None
        bar_width: float, 柱形宽度，默认为0.25
        fig_size: tuple, 图形大小，如(8,6)
        cmap: matplotlib.colors.LinearSegmentedColormap, 颜色映射，默认为grads_default
        font_name: str, 字体名称，默认为'WenQuanYi Zen Hei'
        value_format: str, 数值格式，如'{:.2f}'
        value_range: tuple, Y轴刻度数值范围，如(0,100)
        y_ticks_digits: int, Y轴刻度值的小数位数，默认为1
    }
    示例：
    
    '''       
    x_title= kwargs.get('x_title', x_column)
    y_title= kwargs.get('y_title', None)
    title= kwargs.get('title', None)
    bar_width= kwargs.get('bar_width', 0.25)# 获取柱形的宽度
    fig_size=kwargs.get('fig_size',None)
    cmap=kwargs.get('cmap',cmaps.grads_default)
    font_name= kwargs.get('font_name', 'WenQuanYi Zen Hei') 
    if font_name!=None:
        # 设置中文字体为系统中已安装的字体，如SimSun（宋体）、SimHei（黑体）
        plt.rcParams['font.sans-serif'] = [font_name]  # 设置中文字体为宋体 ['WenQuanYi Zen Hei']
        plt.rcParams['axes.unicode_minus'] = False  # 用于正常显示负号 
    value_format= kwargs.get('value_format', None) 
    value_range= kwargs.get('value_range', (None,None)) 
    y_ticks_digits= kwargs.get('y_ticks_digits', 1)
    # 获取城市列表
    cities = df[x_column]
    if fig_size==None:
        length = len(cities)        
        if length >=8:
            fig_size=(math.ceil(length*0.85), 6)
        else:
            fig_size=(8, 6)                    
    # 设置图形大小
    plt.figure(figsize=fig_size)    
    # 设置柱形的位置
    r1 = range(len(cities))
    postions=[]
    postions.append(r1)    
    for i in range(len(data_columns)):
        r2 = [x + bar_width for x in postions[-1]]
        postions.append(r2)       
    bar_colors=cmap.colors#['b','g','r']
    min_values,max_values=[],[]
    # 绘制柱形图
    for column_name,r,color in zip(data_columns,postions,bar_colors):
        plt.bar(r,  df[column_name], color=color, width=bar_width, label=column_name)   #edgecolor='grey', 
        if not any(value_range):
            max_values.append(df[column_name].max())
            min_values.append(df[column_name].min())
    # 添加标签、标题和图例        
    plt.xlabel(x_title, fontweight='bold')
    plt.xticks([r + bar_width for r in range(len(cities))], cities)
    if y_title!=None:
        plt.ylabel(y_title)
    if title!=None:
        plt.title(title)
    plt.legend()
    min_value,max_value=value_range   
    if not any(value_range):
        min_value,max_value=min(min_values),max(max_values)
        scale_min_factor=0.95 if min_value>=0 else 1.05
        scale_max_factor=1.05 if max_value>=0 else 0.95
        min_value,max_value=min_value*scale_min_factor,max_value*scale_max_factor      
    plt.ylim(min_value,max_value)    
    if value_format is not None:
        interval_y =round(float((max_value - min_value) / 7),y_ticks_digits+1)
        # 设置y轴刻度值和标签
        y_ticks =np.round(np.arange(min_value, max_value, interval_y),y_ticks_digits)
        plt.yticks(y_ticks, [f'{format(value, value_format)}' for value in y_ticks])
    # 显示图形
    plt.show()   
    # 保存图像为图片文件
    plt.savefig(save_file,dpi=300,bbox_inches='tight')
    return save_file


def contour_plot_func(savefile,X, Y, Z,x_name, y_name, cmap='jet'):
    """
    绘制二维等值线图
    """
    # 创建网格
    xx, yy = np.meshgrid(np.arange(0, 1.51, 0.01), np.arange(0, 1.51, 0.01))
    
    # 使用样条插值
    interp_func = interp2d(X, Y, Z, kind='cubic')
    zz = interp_func(xx[0, :], yy[:, 0])

    # 创建图形
    plt.figure(figsize=(10,9))  # 转换为英寸
    plt.title('Total')

    max_z = np.ceil(np.max(zz))
    min_z = np.floor(np.min(zz))    
    color_map_size = np.arange(min_z, max_z + 0.5, 0.5)
    # 找到合适的颜色映射范围
    first_num = max(np.where(min_z >= color_map_size)[0])
    end_num = min(np.where(max_z <= color_map_size)[0])
    v = color_map_size[first_num:end_num + 1]    
    # # 加载颜色映射    
    # 绘制等值线图
    contour = plt.contourf(xx, yy, zz, levels=v, cmap=cmap)
    plt.colorbar(contour)
    
    # 设置轴刻度和标签
    plt.xticks(np.arange(0, 1.51, 0.25))
    plt.yticks(np.arange(0, 1.51, 0.25))
    plt.tick_params(axis='both', which='major', labelsize=22, direction='out')
    
    plt.xlabel(x_name, fontsize=30, fontname='Times New Roman')
    plt.ylabel(y_name, fontsize=30, fontname='Times New Roman')

    # 添加虚线
    plt.plot([1, 1], [0, 1.5], 'k-.', linewidth=1)
    plt.plot([0, 1.5], [1, 1], 'k-.', linewidth=1)

    # 保存图形
    plt.savefig(savefile, bbox_inches='tight', dpi=300)
    plt.close()

if __name__=="__main__":
    print("hello world")    
    
    import pandas as pd
    import numpy as np
    df=pd.read_csv(r"D:\ZQZH.csv")
    # {data_type:{"x":x,"y":y}}
    dict_data={"ZQZH":{"x":df["Obeservation"],"y":df["Monitor"]}}
    draw_multiple_density_plots(dict_data, xName="True Values", yName="Estimated Values", cmap='jet',panel_layout=None,density=True,nbins=200,show_plot=True,show_best_line=False,show_r2=False)
    print("done")
    # 数据准备
    # file_name = 'SD_JAN_PM_ALL'
    # rsm_name = f'pf-ERSM_{file_name}'
    # savefile = 'SaveContourPicture'

    # # 加载数据文件
    # data = loadmat(file_name + '.mat')
    # ersm_data = loadmat(file_name + '_ersm.mat')

    # # 假设这些变量来自加载的数据文件
    # Lpma = ersm_data.get('Lpma')  # 示例
    # Rpma = ersm_data.get('Rpma')  # 示例
    # Ipma = ersm_data.get('Ipma')  # 示例
    # region_cell = ersm_data.get('regionCell')  # 示例

    # matrix_file = 'E:/文档/My RSM-VAT Files/Data/Example/PRD_SD_OCT_41_41/Config Files/Matrix_SD_2015_case358_387_PM.csv'
    # aconc_file = 'E:/文档/My RSM-VAT Files/Data/Example/PRD_SD_OCT_41_41/CMAQ/ACONC.'
    # region_file = 'E:/文档/My RSM-VAT Files/Data/Example/PRD_SD_July_41_41/Config Files/SD_Multi_region_include_other(MATLAB).csv'

    # region_num = 7
    # speces = 4
    # region_row = 148
    # region_col = 112

    #    # Out of Sample
    #     xx = np.arange(0, 1.51, 0.25)
    #     yy = np.arange(0, 1.51, 0.25)

    #     result = np.zeros((len(xx) * len(yy), 1))
    #     zz = np.zeros((len(xx), len(yy)))
    #     x = np.ones((region_num, speces))  # 默认全1基准矩阵
    #     n = 0
    #     x_name = 'SO2'
    #     y_name = 'NH3'
    
    #     for i in range(len(xx)):
    #         for j in range(len(yy)):
    #             x[:, 1] = xx[i]
    #             x[:, 2] = yy[j]
    #             pre_data = get_response_values_pm(Lpma, Rpma, Ipma, region_cell, x, region_row, region_col, region_num, data)       
    #             zz[i, j] =  np.mean(pre_data)
    #             n += 1

    #     contour_plot_func(rsm_name, xx, yy, zz, savefile, x_name, y_name)


    # # 保存结果为 CSV
    # os.makedirs('./SaveCSV', exist_ok=True)
    # np.savetxt(f'./SaveCSV/{rsm_name}_{x_name}{y_name}Total.csv', result, delimiter=',')


import numpy as np


def create_grid_from_receptor_grid_file(file_path, output_path='grid.npy'):
    '''
    创建DeepRSM所需的grid.npy文件
    :param file_path: 基于Receptor grid file（RSM-VAT 中的网格受体文件）进行获取，对同个网格而言，只保留网格占比最大的记录
    :param output_path: 输出文件路径，默认在当前目录下的grid.npy,用户可以自行指定路径
    :return: 没有返回值
    '''
    # 读取文件
    max_ratios = {}
    with open(file_path, 'r') as file:
        for line in file:
            values = line.strip().split()
            id_val = int(values[0])  # ID
            col = int(values[1])  # Col
            row = int(values[2])  # Row
            ratio = float(values[3])  # Ratio
            current_max_ratio, current_id = max_ratios.get((col, row), (float('-inf'), None))
            if ratio > current_max_ratio:
                max_ratios[(col, row)] = (ratio, id_val)

    # 输出结果
    print("ID Col Row Ratio Code City")
    # 获取 Col 和 Row 的最大 Ratio 记录的 ID，并存储到 NumPy 二维数组中
    max_col = max(col for col, _ in max_ratios.keys())
    max_row = max(row for _, row in max_ratios.keys())
    result_array = np.zeros(( max_row,max_col), dtype=int)#为了和后面deeprsm脚本保持一致，这里改为先行后列
    for (col, row), (_, id_val) in max_ratios.items():
        result_array[row - 1][col - 1] = id_val
        # 保存为 .npy 文件
    np.save(output_path, result_array)


if __name__ == "__main__":
    file_path =r'E:\paper_data\PRD_Daily_DeepRSM\Config Files\Model_Grids_Info_PRD.txt'
    create_grid_from_receptor_grid_file(file_path,r'E:\paper_data\PRD_Daily_DeepRSM\Config Files\grid.npy')
    print("Success!")


"""
Author: Devin
Date: 2024-06-18 21:30:32
LastEditors: Devin
LastEditTime: 2024-11-28 17:22:15
Description: 

Copyright (c) 2024 by Devin, All Rights Reserved. 
"""

# from .date_helper import *
# from .earth_helper import *
# from .file_helper import *
# from .log_helper import *
# from .interpolation_helper import *
# from .map_helper import *
# from .netcdf_helper import *
# from .panelayout_helper import *
# from .projection_helper import *
# from .scatter_plot_helper import *
# from  .sqlite_helper import *

import importlib


def lazy_import(module_name):
    module = None

    def _import():
        nonlocal module
        if module is None:
            module = importlib.import_module(f"esil.{module_name}")
        return module

    return _import


date_helper = lazy_import("date_helper")
earth_helper = lazy_import("earth_helper")
file_helper = lazy_import("file_helper")
log_helper = lazy_import("log_helper")
interpolation_helper = lazy_import("interpolation_helper")
map_helper = lazy_import("map_helper")
netcdf_helper = lazy_import("netcdf_helper")
panelayout_helper = lazy_import("panelayout_helper")
projection_helper = lazy_import("projection_helper")
scatter_plot_helper = lazy_import("scatter_plot_helper")
sqlite_helper = lazy_import("sqlite_helper")
wrf_helper = lazy_import("wrf_helper")
regrid_helper = lazy_import("regrid_helper")
plot_helper = lazy_import("plot_helper")
rsm_helper = lazy_import(".rsm_helper/__init__")  # 待测试20240625
